#!/usr/bin/fades --system-site-packages

import datetime
import itertools
import os
import subprocess
import tempfile
from decimal import Decimal

import pyoo  # fades
import matplotlib.pyplot as plt  # fades >=3.9


CONNECT_MSG = """
Need to start Libreffice! Use a command like:
    libreoffice -accept="socket,host=localhost,port=2002;urp;" &
"""

# all useful rows should be less than this
MAX_ROW_DATA = 3000

# epoch
BASE_DATE = datetime.date(1899, 12, 30)

# to round to 4 decimal points
DECIMAL_4_DIGITS = Decimal(".0001")

# the span of info to get
today = datetime.date.today()
LIMIT_DATE_TO = datetime.date(today.year, today.month, 1)
LIMIT_DATE_FROM = datetime.date(today.year - 2, today.month, 1)

# A4 in inches
FIG_WIDTH = 210 / 25.4
FIG_HEIGHT = 297 / 25.4

# configure matplotlib
plt.rcParams.update({'font.size': 5})


class Record:
    """A typical record with expense info."""
    def __init__(self, triplet):
        date_int, self.description, amount_float = triplet
        self.date = datetime.timedelta(days=date_int) + BASE_DATE
        self.amount = Decimal.from_float(amount_float).quantize(DECIMAL_4_DIGITS)

    def __str__(self):
        return "[{:%Y-%m-%d}] {}: {}".format(self.date, self.description, self.amount)


def get_records(sheet, init_col):
    """Get the records from a pack."""
    pack = sheet[0:MAX_ROW_DATA, init_col:init_col + 3].values
    pack = iter(pack)

    # consume the header
    triplet = next(pack)
    assert triplet[1] != ''  # the title
    triplet = next(pack)
    assert triplet[0] == ''
    assert triplet[2] != ''  # the total of the column
    total_amount = Decimal.from_float(triplet[2]).quantize(DECIMAL_4_DIGITS)
    triplet = next(pack)
    assert triplet == ('', '', ''), triplet  # the separator

    # read initial chunk (the visible one where we normally operate)
    initial_chunk = []
    for triplet in pack:
        if triplet[0] == '':
            # end of first chunk
            break
        initial_chunk.append(triplet)

    # consume the middle intermediate space
    for triplet in pack:
        if triplet[0] != '':
            break

    # grab the historic data
    historic = [triplet]
    for triplet in pack:
        if triplet[0] == '':
            break
        historic.append(triplet)
    else:
        raise ValueError("End of data achieved without emptiness, MAX_ROW_DATA not enough")

    # join and process
    all_data = [Record(triplet) for triplet in historic + initial_chunk]
    return total_amount, all_data


def process_pack(sheet, init_col, title):
    """Process a pack of three columns."""
    print("Processing", sheet.name, repr(title))
    total_amount, all_records = get_records(sheet, init_col)

    # validate date ordering
    prev = all_records[0]
    for record in all_records[1:]:
        if record.date < prev.date:
            print("    bad date order!", prev, record)
        prev = record

    records = [r for r in all_records if LIMIT_DATE_FROM <= r.date < LIMIT_DATE_TO]
    rest_amount = sum(r.amount for r in all_records if r.date >= LIMIT_DATE_TO)
    if not records:
        raise ValueError("no records for " + str((sheet, init_col, title)))

    # group by month
    grouped_by_month = {}
    for record in records:
        grouped_by_month.setdefault("{:%Y-%m}".format(record.date), []).append(record)

    # add empty for final/current month (will show the balance here, but no movements)
    grouped_by_month["{:%Y-%m}".format(LIMIT_DATE_TO)] = []

    permonth = []
    for month, records in sorted(grouped_by_month.items()):
        total_income = sum(r.amount for r in records if r.amount > 0)
        total_outcome = sum(r.amount for r in records if r.amount < 0)
        permonth.append((month, total_income, total_outcome))
    return total_amount, rest_amount, permonth


def thousands(value):
    """Divide number by thousands."""
    return int(round(value / 1000)) if value else ""


def calc_limits(valinc, valout):
    raw = int(max(valinc, -valout) / 100)
    round_len = len(str(raw))
    adj = int(round(raw, -(round_len - 1)))

    linc = int(adj * round(valinc) * 1.1 / adj)
    lout = int(adj * round(valout) * 1.3 / adj)  # more margin because label
    return linc, lout


def draw_page(page_name, data):
    all_yearmonths = set()
    for _, _, ym_data in data.values():
        all_yearmonths.update(ym_data.keys())
    all_yearmonths = sorted(all_yearmonths)
    ym_labels = [f"{x[5:]}/{x[2:4]}" for x in all_yearmonths]  # 2024-03 -> 03/24

    titles, all_incomes, all_outcomes, all_limits, all_balances = [], [], [], [], []
    for title, (total_amount, rest_amount, ym_data) in data.items():
        titles.append(title)
        values = [ym_data.get(ym, (0, 0)) for ym in all_yearmonths]
        incomes, outcomes = zip(*values)
        all_incomes.append(incomes)
        all_outcomes.append(outcomes)

        cur_balance = total_amount - rest_amount
        balances = []
        for income, outcome in reversed(list(zip(incomes, outcomes))):
            cur_balance -= (income + outcome)
            balances.append(cur_balance)
        balances.reverse()
        all_balances.append(balances)

        lim_sup = max(itertools.chain(incomes, balances))
        lim_inf = min(itertools.chain((i + o for i, o in zip(incomes, outcomes)), balances))
        all_limits.append((lim_sup, lim_inf))

    # extend the limits so graph is not real close
    all_limits = [calc_limits(i, o) for i, o in all_limits]

    label_locations = list(range(len(all_yearmonths)))
    width = 0.35  # the width of the bars

    fig, all_axis = plt.subplots(len(titles), layout='constrained')
    fig.set_size_inches(FIG_WIDTH, FIG_HEIGHT)
    fig.subplots_adjust(hspace=0.3)

    data = zip(all_axis, all_incomes, all_outcomes, titles, all_limits, all_balances)
    for ax, incomes, outcomes, title, limits, balances in data:
        rects = ax.bar(label_locations, incomes, width)
        rects = ax.bar([x + width for x in label_locations], outcomes, width, bottom=incomes)
        ax.bar_label(rects, padding=5, fmt=thousands)
        ax.set_ylabel(title)
        ax.set_xticks([x + width / 2 for x in label_locations], ym_labels, rotation=90)
        ax.yaxis.set_major_formatter(lambda value, pos: thousands(value))
        ax.set_ylim(bottom=limits[1], top=limits[0])
        ax.axhline(y=0, color='black', linestyle='dotted')
        ax.plot(label_locations, balances, color="red")

    _, fname = tempfile.mkstemp(suffix=".pdf")
    plt.savefig(fname)
    return fname


try:
    desktop = pyoo.Desktop('localhost', 2002)
except IOError:
    print(CONNECT_MSG)
    exit()
docpath = desktop.open_spreadsheet("/home/facundo/Dropbox/Facundo/Distrib.ods")


COLLECT_DATA = {
    "Personas": ["Facundo", "Moni", "Felipe", "Malena", "Vacaciones", "Actividades"],
    "Hogar": ["Casa", "Ahorro general", "Arreglos/mejoras casa", "Empleados", "Salud"],
    "Servicios": [
        "Agua", "Electricidad", "Gas", "Seguro Hogar / Alarma", "Municipalidad", "ARBA", "Cable"],
    "Auto": ["Estac.", "Seguro", "Patente", "Mantenimiento", "Nafta", "Peajes"],
}


all_yms = set()
data = {}
for sheet in docpath.sheets:
    try:
        collect_info = COLLECT_DATA[sheet.name]
    except KeyError:
        continue

    # walk on the titles row, finding useful pack of columns
    cells = sheet[0, 0:100].values
    for column, title in enumerate(cells):
        if title not in collect_info:
            continue
        pivot = column - 1
        total_amount, rest_amount, results = process_pack(sheet, pivot, title)
        for yearmonth, income, outcome in results:
            all_yms.add(yearmonth)
            d = data.setdefault(sheet.name, {})
            _, _, d = d.setdefault(title, (total_amount, rest_amount, {}))
            d[yearmonth] = (income, outcome)

all_pages = []
for pagename, subdata in data.items():
    fname = draw_page(pagename, subdata)
    all_pages.append(fname)

output = "report.pdf"
subprocess.run(["pdftk"] + all_pages + ["cat", "output", output], check=True)
for fname in all_pages:
    os.unlink(fname)
